<?php

namespace Drupal\payment_authnet\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\payment_authnet\Entity\AuthnetProfile;
use Drupal\payment_authnet\Exception;
use Drupal\payment_authnet\Exception\PaymentAuthnetException;
use Drupal\payment_authnet\Exception\PaymentAuthnetApiException;
use net\authorize\api\contract\v1\AuthenticateTestRequest;
use net\authorize\api\controller\AuthenticateTestController;

/**
 * Defines a form to configure maintenance settings for this site.
 */
class TestForm extends FormBase {

  /**
   * The list of Authnet Profiles.
   *
   * @var array
   */
  protected $authnetProfiles;

  /**
   * Constructs a new TestForm object.
   */
  public function __construct() {
    $this->authnetProfiles = AuthnetProfile::loadMultiple();
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'payment_authnet_test';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $options = ['' => $this->t('- Select a profile -')];
    foreach ($this->authnetProfiles as $id => $authnet_profile) {
      $options[$id] = $authnet_profile->label();
    }
    $profile_ids = array_keys($this->authnetProfiles);

    $form['profileId'] = [
      '#title' => $this->t('Authorize.net profile'),
      '#type' => 'select',
      '#options' => $options,
      '#default_value' => count($profile_ids) ? reset($profile_ids) : '',
    ];

    $form['actions'] = [
      'submit' => [
        '#type' => 'submit',
        '#value' => t('Test'),
      ],
    ];

    return $form;
  }

  /**
   * Helper function to send Authorize.net test request to verify credentials.
   *
   * @param string $profileId
   *   Authnet Profile machine name to validate.
   *
   * @return \net\authorize\api\contract\v1\AuthenticateTestResponse
   *   Authorize.net test response result.
   *
   * @throws Drupal\payment_authnet\Exception\PaymentAuthnetApiException
   */
  protected function sendTestRequest($profileId) {
    $profile = $this->authnetProfiles[$profileId];
    // Common setup for API credentials.
    $refId = $profile->getApiKey() . REQUEST_TIME;

    $testRequest = new AuthenticateTestRequest();
    $testRequest->setMerchantAuthentication($profile->getMerchantAuthentication());
    $testRequest->setRefId($refId);

    $controller = new AuthenticateTestController($testRequest);
    $response = $controller->executeWithApiResponse($profile->getApiEnvironment());

    if ($response->getRefId() != $refId) {
      throw new PaymentAuthnetApiException($this->t('Invalid Reference ID in response.'));
    }
    if ($response->getMessages()->getResultCode() != 'Ok') {
      throw new PaymentAuthnetApiException($this->t('Response returned @resultCode. @message (@code)', [
        '@resultCode' => $response->getMessages()->getResultCode(),
        '@message' => $response->getMessages()->getMessage()[0]->getText(),
        '@code' => $response->getMessages()->getMessage()[0]->getCode(),
      ]));
    }
    return $response;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Proper exception handling. Plural messages.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $profileId = $form_state->getValue('profileId');

    if (empty($profileId)) {
      drupal_set_message(t('You need to select a profile to test.'), 'error');
      return;
    }

    try {
      $response = $this->sendTestRequest($profileId);
      drupal_set_message($this->t('Response returned @resultCode. @message (@code).', [
        '@resultCode' => $response->getMessages()->getResultCode(),
        '@message' => $response->getMessages()->getMessage()[0]->getText(),
        '@code' => $response->getMessages()->getMessage()[0]->getCode(),
      ]));
    }
    catch (PaymentAuthnetException $e) {
      drupal_set_message($e->getMessage(), 'error');
    }
    catch (Exception $e) {
      \Drupal::logger('payment_authnet')->error($e->getMessage());
    }

  }

}
