<?php

namespace Drupal\payment_authnet\Plugin\Payment\Method;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Utility\Token;
use Drupal\payment\Entity\PaymentInterface;
use Drupal\payment\EventDispatcherInterface;
use Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodBase;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodCapturePaymentInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodRefundPaymentInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodUpdatePaymentStatusInterface;
use Drupal\payment_authnet\AdditionalFieldsTrait;
use Drupal\payment_authnet\Entity\AuthnetProfile;
use Drupal\payment_authnet\Exception\PaymentAuthnetException;
use Drupal\payment_authnet\Exception\PaymentAuthnetResponseException;
use net\authorize\api\contract\v1\CustomerAddressType;
use net\authorize\api\contract\v1\CreditCardType;
use net\authorize\api\contract\v1\CreateTransactionRequest;
use net\authorize\api\contract\v1\CustomerDataType;
use net\authorize\api\contract\v1\GetTransactionDetailsRequest;
use net\authorize\api\contract\v1\LineItemType;
use net\authorize\api\contract\v1\OrderType;
use net\authorize\api\contract\v1\PaymentType;
use net\authorize\api\contract\v1\TransactionDetailsType;
use net\authorize\api\contract\v1\TransactionRequestType;
use net\authorize\api\contract\v1\UserFieldType;
use net\authorize\api\controller\CreateTransactionController;
use net\authorize\api\controller\GetTransactionDetailsController;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A payment method using Authnet.
 *
 * Plugins that extend this class must have the following keys in their plugin
 * definitions:
 * - entity_id: (string) The ID of the payment method entity the plugin is for.
 * - execute_status_id: (string) The ID of the payment status plugin to set at
 *   payment execution.
 * - capture: (boolean) Whether or not payment capture is supported.
 * - capture_status_id: (string) The ID of the payment status plugin to set at
 *   payment capture.
 * - refund: (boolean) Whether or not payment refunds are supported.
 * - refund_status_id: (string) The ID of the payment status plugin to set at
 *   payment refund.
 *
 * @PaymentMethod(
 *   deriver = "Drupal\payment_authnet\Plugin\Payment\Method\AuthnetDeriver",
 *   id = "payment_authnet",
 *   operations_provider = "\Drupal\payment_authnet\Plugin\Payment\Method\AuthnetOperationsProvider",
 * )
 */
class Authnet extends PaymentMethodBase implements ContainerFactoryPluginInterface, PaymentMethodCapturePaymentInterface, PaymentMethodRefundPaymentInterface, PaymentMethodUpdatePaymentStatusInterface, AuthnetInterface {

  use AdditionalFieldsTrait;

  /**
   * The payment status manager.
   *
   * @var \Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface
   */
  protected $paymentStatusManager;

  /**
   * The AnetAPI payment object (contains info about credit card).
   *
   * @var \net\authorize\api\contract\v1\PaymentType
   */
  protected $AnetApiPayment;

  /**
   * The AnetAPI transaction request object (contains info about transaction).
   *
   * @var \net\authorize\api\contract\v1\TransactionRequestType
   */
  protected $transactionRequest;

  /**
   * The AnetAPI transaction request Type.
   *
   * @var string
   */
  protected $transactionType;

  /**
   * The AnetAPI transaction request Type.
   *
   * @var \net\authorize\api\contract\v1\TransactionDetailsType
   */
  protected $transactionDetails;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Custom Fields Values.
   *
   * @var array
   */
  protected $customFieldsValues;

  /**
   * Authnet Profile.
   *
   * @var \Drupal\payment_authnet\Entity\AuthnetProfileInterface
   */
  protected $AuthnetProfile;

  /**
   * The config factory.
   *
   * Subclasses should use the self::config() method, which may be overridden to
   * address specific needs when loading config, rather than this property
   * directly. See \Drupal\Core\Form\ConfigFormBase::config() for an example of
   * this.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The cache object associated with the 'data' bin.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Constructs a new instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param array $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\payment\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\Utility\Token $token
   *   The token API.
   * @param \Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface $payment_status_manager
   *   The payment status manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache object associated with the 'data' bin.
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, ModuleHandlerInterface $module_handler, EventDispatcherInterface $event_dispatcher, Token $token, PaymentStatusManagerInterface $payment_status_manager, EntityFieldManagerInterface $entity_field_manager, ConfigFactoryInterface $config_factory, CacheBackendInterface $cache) {
    $configuration += $this->defaultConfiguration();
    parent::__construct($configuration, $plugin_id, $plugin_definition, $module_handler, $event_dispatcher, $token, $payment_status_manager);
    $this->entityFieldManager = $entity_field_manager;
    $this->AuthnetProfile = AuthnetProfile::load($this->pluginDefinition['profile']);
    $this->AnetApiPayment = new PaymentType();
    $this->transactionType = PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE;
    $this->refId = $this->AuthnetProfile->getApiKey() . '_' . REQUEST_TIME;
    $this->transactionDetails = new TransactionDetailsType();
    $this->transactionRequest = new TransactionRequestType();
    $this->configFactory = $config_factory;
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('module_handler'),
      $container->get('payment.event_dispatcher'),
      $container->get('token'),
      $container->get('plugin.manager.payment.status'),
      $container->get('entity_field.manager'),
      $container->get('config.factory'),
      $container->get('cache.data')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $elements = parent::buildConfigurationForm($form, $form_state);

    // Add credit card fields.
    $elements['credit_card_number'] = [
      '#type' => 'creditfield_cardnumber',
      '#title' => $this->t('Credit Card Number'),
      '#description' => $this->t('Your 16 digits credit card number.'),
      '#maxlength' => 16,
      '#size' => 16,
      '#required' => TRUE,
    ];

    $elements['expiration_date'] = [
      '#type' => 'creditfield_expiration',
      '#title' => $this->t('Expiration Date'),
      '#required' => TRUE,
    ];

    $elements['credit_card_cvv'] = [
      '#type' => 'creditfield_cardcode',
      '#title' => $this->t('CVV Code'),
      '#maxlength' => 4,
      '#size' => 4,
      '#description' => $this->t('Your 3 or 4 digit security code on the back of your card.'),
      '#required' => TRUE,
    ];
    return $elements + $this->addAdditionalFields();
  }

  /**
   * Creates fields in addition to credit card number, date and cvv.
   *
   * @return array
   *   An array of additional fields to adjust Configuration Form.
   */
  protected function addAdditionalFields() {
    $result = [];
    $field_mapping = $this->configFactory->get('payment_authnet.settings')->get('field_mappings');
    foreach ($this->getAvailableFields() as $section_human_name => $children) {
      $section_name = $this->humanToMachine($section_human_name);
      $result += $this->buildSectionWrapper($section_human_name);
      foreach ($children as $field_name => $type) {
        if (!empty($field_mapping[$section_name][$field_name]['title'])) {
          $result[$section_name][$field_name] = [
            '#type' => $type,
            '#title' => $field_mapping[$section_name][$field_name]['title'],
            '#required' => $field_mapping[$section_name][$field_name]['required'],
            '#default_value' => $this->token->replace($field_mapping[$section_name][$field_name]['default_value']),
          ];
          if ($field_mapping[$section_name][$field_name]['required']) {
            $result[$section_name]['#open'] = TRUE;
            $result[$section_name]['#required'] = TRUE;
          }
        }
      }
    }
    return $result;
  }

  /**
   * Save field value in customFieldValues.
   *
   * @param array $fs_values_field
   *   List of field values.
   * @param string $field_name
   *   Field name.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::submitConfigurationForm()
   */
  protected function setCustomFieldValue(array $fs_values_field, $field_name) {
    if (count($fs_values_field) > 1) {
      foreach ($fs_values_field as $key => $field) {
        $this->customFieldsValues[$field_name][$key] = reset($field);
      }
    }
    else {
      $first_value = reset($fs_values_field);
      $this->customFieldsValues[$field_name] = reset($first_value);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Create the payment data for a credit card.
    $card = $form_state->getValues()['container']['plugin_form'];
    $this->setCreditCard($card['credit_card_number'], $card['expiration_date'], $card['credit_card_cvv']);

    // Save last 4 digits of credit card and expiration date in database.
    // This is useful for refund operations.
    unset($card['credit_card_cvv']);
    $card['credit_card_number'] = substr($card['credit_card_number'], -4);
    $this->setConfiguration($card);

    // Temporarily save other form field values.
    // @todo Find a better way to process form field values.
    // @see https://www.drupal.org/project/payment/issues/2841035
    // @todo Save vaules from address field in appropriate places.
    $fs_values = $form_state->getValues();
    foreach ($this->entityFieldManager->getFieldDefinitions('payment', $this->getPayment()->bundle()) as $field_name => $field_definition) {
      if (!$field_definition->getFieldStorageDefinition()->isBaseField() && isset($fs_values[$field_name])) {
        $this->setCustomFieldValue($fs_values[$field_name], $field_name);
      }
    }
    if (isset($card['billing_information'])) {
      $this->prepareBillingData($card['billing_information']);
    }
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * Prepares customer address type object for AnetApi.
   *
   * @param array $billing_information
   *   Array with Billing information to set.
   *
   * @return \net\authorize\api\contract\v1\CustomerAddressType
   *   Instance of Customer Address Type object.
   */
  protected function prepareBillingData(array $billing_information = []) {
    if (!isset($this->billTo)) {
      $this->billTo = new CustomerAddressType();
    }
    foreach ($billing_information as $key => $value) {
      $method_name = 'set' . str_replace(' ', '', ucwords(str_replace('_', ' ', $key)));
      if (method_exists($this->billTo, $method_name)) {
        $this->billTo->$method_name($value);
      }
    }
    return $this->billTo;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'transactionId' => NULL,
      'credit_card_number' => NULL,
      'expiration_date' => NULL,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedCurrencies() {
    return TRUE;
  }

  /**
   * Gets the status to set on payment execution.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getExecuteStatusId() {
    return $this->pluginDefinition['execute_status_id'];
  }

  /**
   * Gets the status to set on payment capture.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getCaptureStatusId() {
    return $this->pluginDefinition['capture_status_id'];
  }

  /**
   * Gets whether or not capture is enabled.
   *
   * @return bool
   *   Whether or not to support capture.
   */
  public function getCapture() {
    return $this->pluginDefinition['capture'];
  }

  /**
   * Gets the status to set on payment refund.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getRefundStatusId() {
    return $this->pluginDefinition['refund_status_id'];
  }

  /**
   * Gets whether or not refund is enabled.
   *
   * @return bool
   *   Whether or not to support refund.
   */
  public function getRefund() {
    return $this->pluginDefinition['refund'];
  }

  /**
   * Gets Authorize.net PaymentType object.
   *
   * @return \net\authorize\api\contract\v1\PaymentType
   *   Authorize.net paymnet type object.
   */
  protected function getAnetApiPayment() {
    return $this->AnetApiPayment;
  }

  /**
   * Sets Authorize.net Transaction Type.
   *
   * Sets $this->transactionType. Possible values:
   * - PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
   * - PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
   * - PAYMENT_AUTHNET_TRANSACTION_CAPTURE
   * - PAYMENT_AUTHNET_TRANSACTION_REFUND.
   *
   * @param string $operation
   *   Payment Status ID to identify the operation that needs to occur.
   *   Possible values: payment_authorized, payment_success, payment_refunded.
   *
   * @see payment_authnet.module
   * @see https://developer.authorize.net/api/reference/#payment-transactions
   * @see \net\authorize\api\contract\v1\TransactionRequestType::setTransactionType()
   *
   * @return $this
   */
  protected function setTransactionType($operation) {
    $transaction_details = $this->getTransactionDetails();

    // Check if operation accepts money transfer.
    if (in_array($operation, ['payment_success', 'payment_money_transferred'])) {
      $this->transactionType = $transaction_details->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
        ? PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
        : PAYMENT_AUTHNET_TRANSACTION_CAPTURE;
    }

    // Check if status ancestor accepts money transfer.
    $ancestors = $this->getPayment()->getPaymentStatus()->getAncestors();
    if (in_array('payment_money_transferred', $ancestors)) {
      $this->transactionType = $transaction_details->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
        ? PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
        : PAYMENT_AUTHNET_TRANSACTION_CAPTURE;
    }

    // Check if this is refund operation.
    if ('payment_refunded' == $operation
      && in_array($transaction_details->getTransactionType(), [
        'authCaptureTransaction',
        'authOnlyTransaction',
      ])
      && in_array($transaction_details->getTransactionStatus(), [
        'settledSuccessfully',
      ])) {
      $this->transactionType = PAYMENT_AUTHNET_TRANSACTION_REFUND;
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getTransactionType() {
    if (!$this->transactionType) {
      throw new \LogicException($this->t('Transaction type is not set'));
    }
    return $this->transactionType;
  }

  /**
   * Creates the Authorize.net Transaction and sets needed props.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::getAnetApiTransaction()
   *
   * @return $this
   */
  protected function setAnetApiTransactionRequest() {
    $this->transactionRequest->setTransactionType($this->getTransactionType());
    $this->transactionRequest->setAmount($this->getPayment()->getAmount());
    $this->transactionRequest->setOrder($this->prepareAnetApiOrder());
    $this->transactionRequest->setLineItems($this->getAnetApiLineItems());
    $this->transactionRequest->setCustomer($this->prepareAnetApiCustomerData());
    $this->transactionRequest->setBillTo($this->prepareBillingData());
    $transaction_id = $this->getConfiguration('transactionId');
    // For authorized cards we don't need payment info but need Transaction ID.
    if ($transaction_id) {
      $this->transactionRequest->setRefTransId($transaction_id);
      // For refund operations we also need last 4 digits of credit card number
      // and expiration date.
      if ($this->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_REFUND) {
        $credit_card_number = $this->getConfiguration('credit_card_number');
        $expiration_date = $this->getConfiguration('expiration_date');
        $this->setCreditCard($credit_card_number, $expiration_date);
        $this->transactionRequest->setPayment($this->getAnetApiPayment());
      }
    }
    // For new payments set Payment data (authorize or authorize and capture
    // operations).
    else {
      $this->transactionRequest->setPayment($this->getAnetApiPayment());
    }

    return $this;
  }

  /**
   * Creates a new Authorize.net CreditCardType object.
   *
   * Creates and sets a Authorize.net CreditCardType object for
   * $this->AnetApiPayment.
   *
   * @param string $card_number
   *   Credit card number (4 or 20 digits).
   * @param string $expiration_date
   *   Credit card expiration date in format YYYY-MM.
   * @param string $cvv
   *   Optional CVV code (3 or 4 digits) (required for first transactions).
   *
   * @return $this
   */
  protected function setCreditCard($card_number, $expiration_date, $cvv = NULL) {
    $creditCard = new CreditCardType();
    $creditCard->setCardNumber($card_number);
    $creditCard->setExpirationDate($expiration_date);
    $creditCard->setCardCode($cvv);

    $this->AnetApiPayment->setCreditCard($creditCard);

    return $this;
  }

  /**
   * Returns the list of line items for ANetApi Transaction.
   *
   * @return \net\authorize\api\contract\v1\LineItemType[]
   *   Array of objects.
   */
  protected function getAnetApiLineItems() {
    $line_items = [];
    foreach ($this->getPayment()->getLineItems() as $payment_line_item) {
      $line_item = new LineItemType();
      $line_item->setItemId($payment_line_item->getPluginId());
      $line_item->setName(empty($payment_line_item->getName()) ? $payment_line_item->getPluginId() : $payment_line_item->getName());
      $line_item->setQuantity($payment_line_item->getQuantity());
      $line_item->setDescription($payment_line_item->getDescription());
      $line_item->setUnitPrice($payment_line_item->getAmount());
      $line_items[] = $line_item;
    }
    return $line_items;
  }

  /**
   * Prepares the customer's identifying information for Authnet transactions.
   *
   * @return \net\authorize\api\contract\v1\CustomerDataType
   *   Customer Data for Anet Transaction.
   */
  protected function prepareAnetApiCustomerData() {
    $customerData = new CustomerDataType();
    $customerData->setType(PAYMENT_AUTHNET_CUSTOMER_DATA_TYPE);
    $customerData->setId(strval($this->getPayment()->getOwner()->id()));
    $customerData->setEmail($this->getPayment()->getOwner()->getEmail());

    return $customerData;
  }

  /**
   * Creates order information.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::getAnetApiTransactionRequest()
   *
   * @return \net\authorize\api\contract\v1\OrderType
   *   Authnet API Order object.
   */
  protected function prepareAnetApiOrder() {
    $order = new OrderType();
    $order->setInvoiceNumber($this->getPayment()->id());

    return $order;
  }

  /**
   * Creates Authorize.net transaction.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::doExecutePayment()
   *
   * @return \net\authorize\api\contract\v1\TransactionRequestType
   *   Authnet API Transaction Request object.
   */
  protected function getAnetApiTransactionRequest() {
    $this->setAnetApiTransactionRequest();

    // Add some merchant defined fields. These fields won't be stored with
    // the transaction, but will be echoed back in the response.
    // @todo Find the more accurate way to get field values.
    foreach ($this->customFieldsValues as $field_name => $field_value) {
      $merchantDefinedField = new UserFieldType();
      $merchantDefinedField->setName($field_name);
      $merchantDefinedField->setValue(is_string($field_value) ? $field_value : Json::encode($field_value));
      $this->transactionRequest->addToUserFields($merchantDefinedField);
    }
    return $this->transactionRequest;
  }

  /**
   * Assemble the complete transaction request.
   *
   * @return object
   *   Transaction Response.
   *
   * @throws Drupal\payment_authnet\Exception\PaymentAuthnetResponseException
   */
  protected function sendRequest() {
    $request = new CreateTransactionRequest();
    $request->setMerchantAuthentication($this->AuthnetProfile->getMerchantAuthentication());
    $request->setRefId($this->getRefId());
    $request->setTransactionRequest($this->getAnetApiTransactionRequest());
    // Create the controller and get the response.
    $controller = new CreateTransactionController($request);
    $response = $controller->executeWithApiResponse($this->AuthnetProfile->getApiEnvironment());

    if ($response->getRefId() != $this->getRefId()) {
      throw new PaymentAuthnetResponseException($this->t('Invalid Reference ID in response. Error code: @code, Error message: @message', [
        '@code' => $response->getMessages()->getResultCode(),
        '@message' => $response->getMessages()->getMessage()[0]->getText(),
      ]));
    }
    $tresponse = $response->getTransactionResponse();
    // Check to see if the API request was not successful.
    if ($response->getMessages()->getResultCode() != 'Ok') {
      if ($tresponse->getErrors() != NULL) {
        $message = $tresponse->getErrors()[0]->getErrorText();
        $code = $tresponse->getErrors()[0]->getErrorCode();
      }
      else {
        $message = $response->getMessages()->getResultCode() . ': ' . $response->getMessages()->getMessage()[0]->getText();
        $code = 0;
      }
      throw new PaymentAuthnetResponseException($message, $code);
    }

    return $tresponse;
  }

  /**
   * Helper function used for doXXXXPayment() functions.
   *
   * @param string $operation
   *   Payment Status ID to identify the operation that needs to occur.
   *   Possible values: payment_authorized, payment_success, payment_refunded.
   *
   * @return bool
   *   TRUE if operation was successful, FALSE otherwise.
   */
  protected function processResults($operation) {
    $this->setTransactionType($operation);
    try {
      $response = $this->sendRequest();
      // Since the API request was successful, look for a transaction response
      // and parse it to display the results of authorizing the card.
      $message_type = $response->getResponseCode() === '1' ? 'status' : 'warning';
      $messages = $response->getMessages();
      if (count($messages)) {
        drupal_set_message($messages[0]->getDescription(), $message_type);
      }
      else {
        \Drupal::logger('payment_authnet')->log($message_type, print_r($response, 1));
      }

      // Store the Authnet Transaction ID.
      if ($response->getTransId()) {
        $this->setConfiguration(['transactionId' => $response->getTransId()]);
      }
      return TRUE;
    }
    catch (PaymentAuthnetException $e) {
      drupal_set_message($e->getMessage(), 'error');
      \Drupal::logger('payment_authnet')->error($e->getMessage() . ' (' . $e->getCode() . ")|" . $e->getTraceAsString());
    }
    catch (Exception $e) {
      \Drupal::logger('payment_authnet')->error($e->getMessasge());
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getTransactionDetails($refresh = FALSE) {
    $transaction_id = $this->getConfiguration('transactionId');
    if ((!$refresh && $this->transactionDetails->getRefTransId()) || empty($transaction_id)) {
      return $this->transactionDetails;
    }

    $transaction_details = &drupal_static('payment_authnet_transaction_details', []);
    if (!$refresh && isset($transaction_details[$transaction_id])) {
      return $transaction_details[$transaction_id];
    }
    elseif (!$refresh && $cached_data = $this->cache->get('payment_authnet_transaction_details:' . $transaction_id)) {
      $transaction_details[$transaction_id] = $cached_data->data;
      return $transaction_details[$transaction_id];
    }
    $request = new GetTransactionDetailsRequest();
    $request->setMerchantAuthentication($this->AuthnetProfile->getMerchantAuthentication());
    $request->setRefId($this->getRefId());
    $request->setTransId($transaction_id);

    // Create the controller and get the response.
    try {
      $controller = new GetTransactionDetailsController($request);
      $response = $controller->executeWithApiResponse($this->AuthnetProfile->getApiEnvironment());
      if ($response->getRefId() != $this->getRefId()) {
        $errorMessages = $response->getMessages()->getMessage();
        throw new PaymentAuthnetResponseException($this->t('Authorize.net error for payment @payment_id: Invalid Reference ID in response. @code: @message (@status)', [
          '@payment_id' => $this->getPayment()->id(),
          '@code' => $response->getMessages()->getResultCode(),
          '@message' => $response->getMessages()->getMessage()[0]->getText(),
          '@status' => $errorMessages[0]->getCode(),
        ]));
      }
      if ($response->getMessages()->getResultCode() != 'Ok') {
        $errorMessages = $response->getMessages()->getMessage();
        throw new PaymentAuthnetResponseException($this->t('Authorize.net error for payment @payment_id: @message (@status)', [
          '@payment_id' => $this->getPayment()->id(),
          '@message' => $errorMessages[0]->getText(),
          '@status' => $errorMessages[0]->getCode(),
        ]));
      }

      $this->transactionDetails = $response->getTransaction();
    }
    catch (PaymentAuthnetResponseException $e) {
      drupal_set_message($e->getMessage(), 'error', FALSE);
    }

    $transaction_details[$transaction_id] = $this->transactionDetails;
    $this->cache->set('payment_authnet_transaction_details:' . $transaction_id, $this->transactionDetails, Cache::PERMANENT, ['payment:' . $this->getPayment()->id()]);
    return $this->transactionDetails;
  }

  /**
   * {@inheritdoc}
   */
  protected function doExecutePayment() {
    if ($this->processResults($this->getExecuteStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getExecuteStatusId()));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doCapturePayment() {
    if ($this->processResults($this->getCaptureStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getCaptureStatusId()));
      $this->getPayment()->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doCapturePaymentAccess(AccountInterface $account) {
    return $this->getCapture()
    && $this->getPayment()->getPaymentStatus()->getPluginId() != $this->getCaptureStatusId();
  }

  /**
   * {@inheritdoc}
   */
  public function doRefundPayment() {
    if ($this->processResults($this->getRefundStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getRefundStatusId()));
      $this->getPayment()->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doRefundPaymentAccess(AccountInterface $account) {
    return
      $this->getRefund()
      && $this->getPayment()->getPaymentStatus()->getPluginId() != $this->getRefundStatusId()
      && $this->getConfiguration('credit_card_number')
      && $this->getConfiguration('expiration_date')
      && $this->getConfiguration('transactionId')
      && $this->getTransactionDetails()->getTransactionStatus() == 'settledSuccessfully';
  }

  /**
   * {@inheritdoc}
   */
  public function updatePaymentStatusAccess(AccountInterface $account) {
    return AccessResult::allowed();
  }

  /**
   * {@inheritdoc}
   */
  public function getSettablePaymentStatuses(AccountInterface $account, PaymentInterface $payment) {
    return [
      'payment_pending',
      'payment_success',
      'payment_authorized',
      'payment_refunded',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration($key = NULL) {
    if ($key) {
      return empty($this->configuration[$key]) ? NULL : $this->configuration[$key];
    }
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function getRefId() {
    return $this->refId;
  }

}
