<?php

namespace Drupal\payment_authnet\Plugin\views\field;

use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\field\FieldPluginBase;
use Drupal\views\ResultRow;
use Drupal\payment_authnet\Plugin\Payment\Method\AuthnetInterface;
use net\authorize\api\contract\v1\TransactionDetailsType;

/**
 * Field handler to get the transaction details for Payment Authnet method.
 *
 * @ingroup views_field_handlers
 *
 * @ViewsField("payment_authnet_transaction_details")
 */
class PaymentAuthnetTransactionDetails extends FieldPluginBase {

  /**
   * {@inheritdoc}
   */
  public function query() {
    // Leave empty to avoid a query on this field.
  }

  /**
   * Define the available options.
   *
   * @return array
   *   An array with options and default values.
   */
  protected function defineOptions() {
    $options = parent::defineOptions();
    $options['payment_authnet_display'] = ['default' => 'card_type'];

    return $options;
  }

  /**
   * Provide the options form.
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    $form['payment_authnet_display'] = [
      '#type' => 'select',
      '#title' => $this->t('Authorize.net transaction details to display'),
      '#default_value' => $this->options['payment_authnet_display'],
      '#description' => $this->t('Select What information you would like to display'),
      '#options' => [
        'card_type' => $this->t('Card Type'),
        'card_code_response' => $this->t('Card code verification (CCV) response'),
        'transaction_status' => $this->t('The status of the transaction'),
        'transaction_type' => $this->t('The type of transaction that was originally submitted'),
        'response_code' => $this->t('The overall status of the transaction'),
        'customer_iP' => $this->t('The IPv4 address of the customer initiating the transaction'),
        'cAVV_response' => $this->t('Cardholder authentication verification response code'),
      ],
    ];
    parent::buildOptionsForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function render(ResultRow $values) {
    /** @var Drupal\payment\Entity\PaymentInterface $payment */
    $payment = $this->getEntity($values);
    $payment_method = $payment->getPaymentMethod();
    $result = '';
    if ($payment_method instanceof AuthnetInterface) {
      $result = $this->getValueFromTransactionDetails($payment_method->getTransactionDetails());
    }

    return $result;
  }

  /**
   * Search value in the Transaction details response.
   *
   * @param \net\authorize\api\contract\v1\TransactionDetailsType $transaction_details
   *   Instance of Authnet API transaction Details Type.
   *
   * @return string|\Drupal\Core\StringTranslation\TranslatableMarkup
   *   Found value or empty.
   */
  protected function getValueFromTransactionDetails(TransactionDetailsType $transaction_details) {
    $result = '';
    $display = $this->options['payment_authnet_display'];

    switch ($display) {
      case 'card_type':
        if ($transaction_details->getPayment()) {
          $result = $transaction_details->getPayment()->getCreditCard()->getCardType();
        }
        break;

      default:
        $this->callDisplayMethodIfFound('get', $transaction_details, $display, $result);
        $this->callDisplayMethodIfFound('format', $this, $display, $result);
        break;
    }

    return $result;
  }

  /**
   * Call getter / formatter for selected transaction details property.
   *
   * @param string $prefix
   *   Method name prefix ('get' or 'format').
   * @param object $object
   *   Object to search method in.
   * @param string $display
   *   Display (property) to search get/format method for.
   * @param string &$result
   *   Result.
   */
  protected function callDisplayMethodIfFound($prefix, $object, $display, &$result) {
    $method_name = $prefix . str_replace(' ', '', ucwords(str_replace('_', ' ', $display)));
    if (method_exists($object, $method_name)) {
      $result = call_user_func([$object, $method_name], $result);
    }
  }

  /**
   * Convert Transaction CAVV response code into human readable string.
   *
   * @param string $code
   *   Cardholder authentication verification response code.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   Translated string with code explanation.
   */
  protected function formatCAVVResponse($code = NULL) {
    $statuses = [
      '0' => $this->t('CAVV was not validated because erroneous data was submitted.'),
      '1' => $this->t('CAVV failed validation.'),
      '2' => $this->t('CAVV passed validation.'),
      '3' => $this->t('CAVV validation could not be performed; issuer attempt incomplete.'),
      '4' => $this->t('CAVV validation could not be performed; issuer system error.'),
      '5' => $this->t('Reserved for future use.'),
      '6' => $this->t('Reserved for future use.'),
      '7' => $this->t('CAVV failed validation, but the issuer is available. Valid for U.S.-issued card submitted to non-U.S acquirer.'),
      '8' => $this->t('CAVV passed validation and the issuer is available. Valid for U.S.-issued card submitted to non-U.S. acquirer.'),
      '9' => $this->t('CAVV failed validation, but the issuer is available. Valid for U.S.-issued card submitted to non-U.S acquirer.'),
      'A' => $this->t('CAVV passed validation but the issuer unavailable. Valid for U.S.-issued card submitted to non-U.S acquirer.'),
      'B' => $this->t('CAVV passed validation, information only, no liability shift.'),
    ];
    if (isset($statuses[(string) $code])) {
      return $statuses[(string) $code];
    }
    return $this->t('CAVV not validated');
  }

  /**
   * Convert Transaction response code into human readable string.
   *
   * @param string $code
   *   Response code (1..4)
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   Translated string with code explanation.
   */
  protected function formatResponseCode($code = NULL) {
    $response_codes = [
      '1' => $this->t('Approved'),
      '2' => $this->t('Declined'),
      '3' => $this->t('Error'),
      '4' => $this->t('Held for Review'),
    ];
    if (isset($response_codes[(string) $code])) {
      return $response_codes[(string) $code];
    }
    return $this->t('Unknown');
  }

  /**
   * Convert Transaction CCV code into human readable string.
   *
   * @param string $code
   *   One letter Card Code Verification Response.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   Translated string with code explanation.
   */
  protected function formatCardCodeResponse($code = NULL) {
    $card_code_responses = [
      'M' => $this->t('CVV matched.'),
      'N' => $this->t('CVV did not match.'),
      'P' => $this->t('CVV was not processed.'),
      'S' => $this->t('CVV should have been present but was not indicated.'),
      'U' => $this->t('The issuer was unable to process the CVV check.'),
    ];
    if (isset($card_code_responses[(string) $code])) {
      return $card_code_responses[(string) $code];
    }
    return $this->t('Unknown');
  }

}
